@@ -1,7 +1,9 @@
 /* NickServ functions.
  *
- * Services is copyright (c) 1996-1999 Andy Church.
+ * Services is copyright (c) 1996-1999 Andrew Church.
  *     E-mail: <achurch@dragonfire.net>
+ * Services is copyright (c) 1999-2000 Andrew Kempe.
+ *     E-mail: <theshadow@shadowfire.org>
  * This program is free but copyrighted software; see the file COPYING for
  * details.
  */
@@ -10,16 +12,16 @@
  * Note that with the addition of nick links, most functions in Services
  * access the "effective nick" (u->ni) to determine privileges and such.
  * The only functions which access the "real nick" (u->real_ni) are:
- *	various functions which set/check validation flags
- *	    (validate_user, cancel_user, nick_{identified,recognized},
- *	     do_identify)
- *	validate_user (adding a collide timeout on the real nick)
- *	cancel_user (deleting a timeout on the real nick)
- *	do_register (checking whether the real nick is registered)
- *	do_drop (dropping the real nick)
- *	do_link (linking the real nick to another)
- *	do_unlink (unlinking the real nick)
- *	chanserv.c/do_register (setting the founder to the real nick)
+ *      various functions which set/check validation flags
+ *          (validate_user, cancel_user, nick_{identified,recognized},
+ *           do_identify)
+ *      validate_user (adding a collide timeout on the real nick)
+ *      cancel_user (deleting a timeout on the real nick)
+ *      do_register (checking whether the real nick is registered)
+ *      do_drop (dropping the real nick)
+ *      do_link (linking the real nick to another)
+ *      do_unlink (unlinking the real nick)
+ *      chanserv.c/do_register (setting the founder to the real nick)
  * plus a few functions in users.c relating to nick creation/changing.
  */
 
@@ -29,12 +31,19 @@
 /*************************************************************************/
 
 static NickInfo *nicklists[256];	/* One for each initial character */
+static NickSuspend *nicksuspends;	/* First suspended nick */
 
 #define TO_COLLIDE   0			/* Collide the user with this nick */
 #define TO_RELEASE   1			/* Release a collided nick */
 
 /*************************************************************************/
 
+static NickSuspend *find_nicksuspend(const char *nick);
+static NickSuspend *add_nicksuspend(NickInfo *ni, const char *reason,
+		const char *who, const time_t expires);
+static void del_nicksuspend(NickSuspend *snick);
+static void unsuspend(NickInfo *ni);
+
 static int is_on_access(User *u, NickInfo *ni);
 static void alpha_insert_nick(NickInfo *ni);
 static NickInfo *makenick(const char *nick);
@@ -47,6 +56,9 @@ static void release(NickInfo *ni, int from_timeout);
 static void add_ns_timeout(NickInfo *ni, int type, time_t delay);
 static void del_ns_timeout(NickInfo *ni, int type);
 
+#ifdef DEBUG_COMMANDS
+static void do_commands(User *u);
+#endif
 static void do_help(User *u);
 static void do_register(User *u);
 static void do_identify(User *u);
@@ -73,22 +85,30 @@ static void do_ghost(User *u);
 static void do_status(User *u);
 static void do_getpass(User *u);
 static void do_forbid(User *u);
+static void do_suspend(User *u);
+static void do_unsuspend(User *u);
 
 /*************************************************************************/
 
 static Command cmds[] = {
+#ifdef DEBUG_COMMANDS
+    { "COMMANDS", do_commands, NULL,  -1, 		      -1,-1,-1,-1 },
+#endif
     { "HELP",     do_help,     NULL,  -1,                     -1,-1,-1,-1 },
     { "REGISTER", do_register, NULL,  NICK_HELP_REGISTER,     -1,-1,-1,-1 },
     { "IDENTIFY", do_identify, NULL,  NICK_HELP_IDENTIFY,     -1,-1,-1,-1 },
+    { "SIDENTIFY",do_identify, NULL,  -1,     -1,-1,-1,-1 },
     { "DROP",     do_drop,     NULL,  -1,
 		NICK_HELP_DROP, NICK_SERVADMIN_HELP_DROP,
 		NICK_SERVADMIN_HELP_DROP, NICK_SERVADMIN_HELP_DROP },
     { "ACCESS",   do_access,   NULL,  NICK_HELP_ACCESS,       -1,-1,-1,-1 },
     { "LINK",     do_link,     NULL,  NICK_HELP_LINK,         -1,-1,-1,-1 },
-    { "UNLINK",   do_unlink,   NULL,  NICK_HELP_UNLINK,       -1,-1,-1,-1 },
+    { "UNLINK",   do_unlink,   NULL,  NICK_HELP_UNLINK,       
+    		-1, NICK_SERVADMIN_HELP_UNLINK, NICK_SERVADMIN_HELP_UNLINK,
+		NICK_SERVADMIN_HELP_UNLINK },
     { "SET",      do_set,      NULL,  NICK_HELP_SET,
-		-1, NICK_SERVADMIN_HELP_SET,
-		NICK_SERVADMIN_HELP_SET, NICK_SERVADMIN_HELP_SET },
+		-1, NICK_SERVADMIN_HELP_SET, NICK_SERVADMIN_HELP_SET, 
+		NICK_SERVADMIN_HELP_SET },
     { "SET PASSWORD", NULL,    NULL,  NICK_HELP_SET_PASSWORD, -1,-1,-1,-1 },
     { "SET URL",      NULL,    NULL,  NICK_HELP_SET_URL,      -1,-1,-1,-1 },
     { "SET EMAIL",    NULL,    NULL,  NICK_HELP_SET_EMAIL,    -1,-1,-1,-1 },
@@ -119,6 +139,12 @@ static Command cmds[] = {
     { "FORBID",   do_forbid,   is_services_admin,  -1,
 		-1, NICK_SERVADMIN_HELP_FORBID,
 		NICK_SERVADMIN_HELP_FORBID, NICK_SERVADMIN_HELP_FORBID },
+    { "SUSPEND", do_suspend,   is_services_admin, -1,
+		-1, NICK_SERVADMIN_HELP_SUSPEND, 
+		NICK_SERVADMIN_HELP_SUSPEND, NICK_SERVADMIN_HELP_SUSPEND },
+    { "UNSUSPEND", do_unsuspend,   is_services_admin, -1,
+		-1, NICK_SERVADMIN_HELP_UNSUSPEND, 
+		NICK_SERVADMIN_HELP_UNSUSPEND, NICK_SERVADMIN_HELP_UNSUSPEND },
     { NULL }
 };
 
@@ -151,7 +177,7 @@ void listnicks(int count_only, const char *nick)
 	struct tm *tm;
 	char buf[512];
 	static const char commastr[] = ", ";
-	int need_comma = 1;
+	int need_comma = 0;
 
 	if (!(ni = findnick(nick))) {
 	    printf("%s not registered.\n", nick);
@@ -163,10 +189,10 @@ void listnicks(int count_only, const char *nick)
 	printf("%s is %s\n", nick, ni->last_realname);
 	printf("Last seen address: %s\n", ni->last_usermask);
 	tm = localtime(&ni->time_registered);
-	strftime(buf, sizeof(buf), getstring(NULL,STRFTIME_DATE_TIME_FORMAT), tm);
+	strftime(buf, sizeof(buf), getstring((NickInfo *)NULL,STRFTIME_DATE_TIME_FORMAT), tm);
 	printf("  Time registered: %s\n", buf);
 	tm = localtime(&ni->last_seen);
-	strftime(buf, sizeof(buf), getstring(NULL,STRFTIME_DATE_TIME_FORMAT), tm);
+	strftime(buf, sizeof(buf), getstring((NickInfo *)NULL,STRFTIME_DATE_TIME_FORMAT), tm);
 	printf("   Last seen time: %s\n", buf);
 	if (ni->url)
 	    printf("              URL: %s\n", ni->url);
@@ -188,13 +214,21 @@ void listnicks(int count_only, const char *nick)
 			need_comma ? commastr : "");
 	    need_comma = 1;
 	}
+	if (ni->status & NS_NO_EXPIRE) {
+	    end += snprintf(buf, sizeof(buf)-(end-buf), "%sNo Expire",
+			need_comma ? commastr : "");
+	    need_comma = 1;
+	}
 	printf("          Options: %s\n", *buf ? buf : "None");
 
     } else {
 
 	for (i = 0; i < 256; i++) {
 	    for (ni = nicklists[i]; ni; ni = ni->next) {
-		printf("    %-20s  %s\n", ni->nick, ni->last_usermask);
+		printf("    %s %-20s  %s\n", 
+			    ni->status & NS_NO_EXPIRE ? "!" : " ", 
+			    ni->nick, ni->status & NS_VERBOTEN ? 
+				     "Disallowed (FORBID)" : ni->last_usermask);
 		count++;
 	    }
 	}
@@ -433,6 +467,7 @@ void load_ns_dbase(void)
 
     switch (ver = get_file_version(f)) {
 
+      case 8:
       case 7:
       case 6:
       case 5:
@@ -504,6 +539,16 @@ void load_ns_dbase(void)
 		    SAFE(read_int32(&ni->flags, f));
 		    if (!NSAllowKillImmed)
 			ni->flags &= ~NI_KILL_IMMED;
+		    if (ni->flags & NI_SUSPENDED) {
+			NickSuspend *snick;
+			snick = add_nicksuspend(ni, "\0", "\0", 0);
+			SAFE(read_buffer(snick->suspendinfo.who, f));
+			SAFE(read_string(&snick->suspendinfo.reason, f));
+			SAFE(read_int32(&tmp32, f));
+			snick->suspendinfo.suspended = tmp32;
+			SAFE(read_int32(&tmp32, f));
+			snick->suspendinfo.expires = tmp32;
+		    }
 		    SAFE(read_int16(&ni->accesscount, f));
 		    if (ni->accesscount) {
 			char **access;
@@ -616,6 +661,17 @@ void save_ns_dbase(void)
 		SAFE(write_string(NULL, f));
 		SAFE(write_int16(ni->linkcount, f));
 		SAFE(write_int32(ni->flags, f));
+		if (ni->flags & NI_SUSPENDED) {
+		    NickSuspend *snick = find_nicksuspend(ni->nick);
+		    if (!snick)	/* for the moment, just in case! */
+			snick = add_nicksuspend(ni, 
+				"<unknown reason; automatically generated>", 
+				"<unknown>", 0);
+		    SAFE(write_buffer(snick->suspendinfo.who, f));
+		    SAFE(write_string(snick->suspendinfo.reason, f));
+		    SAFE(write_int32(snick->suspendinfo.suspended, f));
+		    SAFE(write_int32(snick->suspendinfo.expires, f));
+		}
 		SAFE(write_int16(ni->accesscount, f));
 		for (j=0, access=ni->access; j<ni->accesscount; j++, access++)
 		    SAFE(write_string(*access, f));
@@ -660,9 +716,9 @@ int validate_user(User *u)
     if (!(ni = u->real_ni))
 	return 0;
 
-    if (ni->status & NS_VERBOTEN) {
+    if ((ni->status & NS_VERBOTEN) || (u->ni && (u->ni->flags & NI_SUSPENDED))){
 	notice_lang(s_NickServ, u, NICK_MAY_NOT_BE_USED);
-#ifdef IRC_DAL4_4_15
+#if defined(IRC_DAL4_4_15) || defined(IRC_BAHAMUT)
 	if (NSForceNickChange)
 	    notice_lang(s_NickServ, u, FORCENICKCHANGE_IN_1_MINUTE);
 	else
@@ -672,6 +728,7 @@ int validate_user(User *u)
 	return 0;
     }
 
+#ifndef IRC_BAHAMUT
     if (!NoSplitRecovery) {
 	/* XXX: This code should be checked to ensure it can't be fooled */
 	if (ni->id_timestamp != 0 && u->signon == ni->id_timestamp) {
@@ -683,6 +740,7 @@ int validate_user(User *u)
 	    }
 	}
     }
+#endif
 
     on_access = is_on_access(u, u->ni);
     if (on_access)
@@ -712,7 +770,7 @@ int validate_user(User *u)
 	if (u->ni->flags & NI_KILL_IMMED) {
 	    collide(ni, 0);
 	} else if (u->ni->flags & NI_KILL_QUICK) {
-#ifdef IRC_DAL4_4_15
+#if defined(IRC_DAL4_4_15) || defined(IRC_BAHAMUT)
 	    if (NSForceNickChange)
 	    	notice_lang(s_NickServ, u, FORCENICKCHANGE_IN_20_SECONDS);
 	    else
@@ -720,7 +778,7 @@ int validate_user(User *u)
 	    	notice_lang(s_NickServ, u, DISCONNECT_IN_20_SECONDS);
 	    add_ns_timeout(ni, TO_COLLIDE, 20);
 	} else {
-#ifdef IRC_DAL4_4_15
+#if defined(IRC_DAL4_4_15) || defined(IRC_BAHAMUT)
 	    if (NSForceNickChange)
 	    	notice_lang(s_NickServ, u, FORCENICKCHANGE_IN_1_MINUTE);
 	    else
@@ -741,20 +799,22 @@ int validate_user(User *u)
 void cancel_user(User *u)
 {
     NickInfo *ni = u->real_ni;
-    if (ni) {
+    char realname[NICKMAX+16]; /*Long enough for s_NickServ and " Enforcement"*/
 
-#ifdef IRC_DAL4_4_15
+    if (ni) {
+	snprintf(realname, sizeof(realname), "%s Enforcement", s_NickServ);
+
+#if defined(IRC_DAL4_4_15) || defined(IRC_BAHAMUT)
 	if (ni->status & NS_GUESTED) {
-	    send_cmd(NULL, "NICK %s %ld 1 %s %s %s :%s Enforcement",
-			u->nick, time(NULL), NSEnforcerUser, NSEnforcerHost, 
-			ServerName, s_NickServ);
+	    send_nick(u->nick, NSEnforcerUser, NSEnforcerHost, ServerName,
+	    		realname);
 	    add_ns_timeout(ni, TO_RELEASE, NSReleaseTimeout);
 	    ni->status &= ~NS_TEMPORARY;
 	    ni->status |= NS_KILL_HELD;
 	} else {
 #endif
 	    ni->status &= ~NS_TEMPORARY;
-#ifdef IRC_DAL4_4_15
+#if defined(IRC_DAL4_4_15) || defined(IRC_BAHAMUT)
 	}
 #endif
 	del_ns_timeout(ni, TO_COLLIDE);
@@ -781,6 +841,25 @@ int nick_recognized(User *u)
 
 /*************************************************************************/
 
+/* Remove nickname suspensions that have expired. */
+
+void expire_nicksuspends(void)
+{
+    NickSuspend *ns, *next;
+    time_t now = time(NULL);
+
+    for (ns = nicksuspends; ns; ns = next) {
+	next = ns->next;
+	if (ns->suspendinfo.expires != 0 && ns->suspendinfo.expires <= now) {
+	    log("Expiring nick-suspend for %s", ns->ni->nick);
+	    unsuspend(ns->ni);
+	    del_nicksuspend(ns);
+	}
+    }
+}
+
+/*************************************************************************/
+
 /* Remove all nicks which have expired.  Also update last-seen time for all
  * nicks.
  */
@@ -801,13 +880,14 @@ void expire_nicks()
 	    u->real_ni->last_seen = time(NULL);
 	}
     }
-    if (!NSExpire)
+    if (!NSExpire || opt_noexpire)
 	return;
     for (i = 0; i < 256; i++) {
 	for (ni = nicklists[i]; ni; ni = next) {
 	    next = ni->next;
 	    if (now - ni->last_seen >= NSExpire
-			&& !(ni->status & (NS_VERBOTEN | NS_NO_EXPIRE))) {
+			&& !((ni->status & (NS_VERBOTEN | NS_NO_EXPIRE)) 
+				|| (ni->flags & NI_SUSPENDED))) {
 		log("Expiring nickname %s", ni->nick);
 		delnick(ni);
 	    }
@@ -865,6 +945,92 @@ NickInfo *getlink(NickInfo *ni)
 /*********************** NickServ private routines ***********************/
 /*************************************************************************/
 
+/* Return the structure containing the suspension info for the given nick.
+ * If nick can't be found, returns NULL.
+ */
+
+static NickSuspend *find_nicksuspend(const char *nick)
+{
+    NickSuspend *snick;
+
+    for (snick = nicksuspends; snick; snick = snick->next)
+	if (stricmp(snick->ni->nick, nick) == 0)
+	    return snick;
+
+    return NULL;
+}
+
+/* Insert a new NickSuspend info structure into the suspended nick list and
+ * fill in the values. 
+ */
+
+static NickSuspend *add_nicksuspend(NickInfo *ni, const char *reason,
+		const char *who, const time_t expires)
+{
+    NickSuspend *snick;
+
+    snick = scalloc(sizeof(NickSuspend), 1);
+    snick->ni = ni;
+    strscpy(snick->suspendinfo.who, who, NICKMAX);	
+    snick->suspendinfo.reason = sstrdup(reason);
+    snick->suspendinfo.suspended = time(NULL);
+    snick->suspendinfo.expires = expires;
+
+    if (nicksuspends)
+	nicksuspends->prev = snick;
+    snick->prev = NULL;
+    snick->next = nicksuspends;
+    nicksuspends = snick;
+
+    return snick;
+}
+
+/* Delete the structure containing the suspension info for the given nick. */
+/* FIXME: Possible redundant debug log code below. */
+
+static void del_nicksuspend(NickSuspend *snick)
+{
+    if (snick) {
+	if (debug)
+	    log("debug: del_nicksuspend called for %s", snick->ni->nick);
+	free(snick->suspendinfo.reason);
+	if (snick->prev)
+	    snick->prev->next = snick->next;
+	else
+	    nicksuspends = snick->next;
+	if (snick->next)
+	    snick->next->prev = snick->prev;
+	free(snick);
+	if (debug)
+	    log("debug: del_nicksuspend completed.");
+    }
+}
+
+/* Do everything nescessary to unsuspend the given nickname and remove the
+ * suspension information. We also alter the last_seen value to ensure that
+ * it does not expire within the next NSSuspendGrace seconds; giving the
+ * owner a chance to reclaim it.
+ */
+/* FIXME: Possible redundant debug log code below. */
+
+static void unsuspend(NickInfo *ni)
+{
+    time_t now = time(NULL);
+
+    if (debug)
+	log("debug: unsuspend called for %s", ni->nick);
+    ni->flags &= ~NI_SUSPENDED;
+    if (NSExpire && NSSuspendGrace && 
+			(now - ni->last_seen >= NSExpire - NSSuspendGrace)) {
+	ni->last_seen = now - NSExpire + NSSuspendGrace; 
+	log("Altering last_seen time for %s to %ld.", ni->nick, ni->last_seen);
+    }
+    if (debug)
+	log("debug: unsuspend completed.");
+} 
+
+/*************************************************************************/
+
 /* Is the given user's address on the given nick's access list?  Return 1
  * if so, 0 if not. */
 
@@ -931,15 +1097,22 @@ static NickInfo *makenick(const char *nick)
 
 /*************************************************************************/
 
-/* Remove a nick from the NickServ database.  Return 1 on success, 0
- * otherwise.  Also deletes the nick from any ChanServ/OperServ lists it is
- * on.
+/* Remove a nick from the NickServ database. Return 1 on success, 0
+ * otherwise. If the nick is suspended, delete the suspend info and then 
+ * remove the nick from any ChanServ/OperServ lists it is on.
  */
 
 static int delnick(NickInfo *ni)
 {
+    NickSuspend *ns;
     int i;
 
+    if (ni->flags & NI_SUSPENDED) {
+	if (!(ns = find_nicksuspend(ni->nick)))
+	    log("delnick: couldn't find suspend info for %s", ni->nick);
+	else
+	    del_nicksuspend(ns);
+    }
     cs_remove_nick(ni);
     os_remove_nick(ni);
     if (ni->linkcount)
@@ -1040,44 +1213,70 @@ static void delink(NickInfo *ni)
 /* Collide a nick. 
  *
  * When connected to a network using DALnet servers, version 4.4.15 and above, 
- * Services is now able to force a nick change instead of killing the user. 
+ * Services is able to force a nick change instead of killing the user. 
  * The new nick takes the form "Guest######". If a nick change is forced, we
  * do not introduce the enforcer nick until the user's nick actually changes. 
- * This is watched for and done in cancel_user(). -TheShadow 
+ * This is watched for and done in cancel_user(). 
+ * 
+ * The number that is used to build the "Guest" nick should be:
+ * 	- Atomic
+ * 	- Unique within a 24 hour period - basically this means that if someone
+ * 	  has a "Guest" nick for more than 24 hours, there is a slight chance
+ * 	  that a new "Guest" could be collided. I really don't see this as a 
+ * 	  problem.
+ * 	- Reasonably unpredictable
+ * 
+ * -TheShadow 
  */
 
 static void collide(NickInfo *ni, int from_timeout)
 {
     User *u;
+    char realname[NICKMAX+16]; /*Long enough for s_NickServ and " Enforcement"*/
 
     u = finduser(ni->nick);
 
     if (!from_timeout)
 	del_ns_timeout(ni, TO_COLLIDE);
 
-#ifdef IRC_DAL4_4_15
+#if defined(IRC_DAL4_4_15) || defined(IRC_BAHAMUT)
     if (NSForceNickChange) {
 	struct timeval tv;
 	char guestnick[NICKMAX];
+	static long sec = -1;
+	static long usec = -1;
+	long tmp_sec, tmp_usec;
+	uint16 counter = 0;
 
 	gettimeofday(&tv, NULL);
-	snprintf(guestnick, sizeof(guestnick), "%s%ld%ld", NSGuestNickPrefix,
-			tv.tv_usec / 10000, tv.tv_sec % (60*60*24));
+	tmp_sec = tv.tv_sec % (60*60*24);
+	tmp_usec = tv.tv_usec / 10000;
+	if (usec == tmp_usec && sec == tmp_sec) {
+	    counter++;
+	} else {
+	    sec = tmp_sec;
+	    usec = tmp_usec;
+	    counter = 0;
+	}
+	snprintf(guestnick, sizeof(guestnick), "%s%ld%d%ld", 
+			NSGuestNickPrefix, usec, counter, sec);
 
         notice_lang(s_NickServ, u, FORCENICKCHANGE_NOW, guestnick);
-
-	send_cmd(NULL, "SVSNICK %s %s :%ld", ni->nick, guestnick, time(NULL));
+	send_cmd(s_NickServ, "SVSNICK %s %s :%ld", 
+			ni->nick, guestnick, time(NULL));
 	ni->status |= NS_GUESTED;
     } else {
 #endif
 	notice_lang(s_NickServ, u, DISCONNECT_NOW);
     	kill_user(s_NickServ, ni->nick, "Nick kill enforced");
-    	send_cmd(NULL, "NICK %s %ld 1 %s %s %s :%s Enforcement",
-		ni->nick, time(NULL), NSEnforcerUser, NSEnforcerHost,
-		ServerName, s_NickServ);
+
+	snprintf(realname, sizeof(realname), "%s Enforcement", s_NickServ);
+	send_nick(ni->nick, NSEnforcerUser, NSEnforcerHost, ServerName,
+		realname);
+
 	ni->status |= NS_KILL_HELD;
 	add_ns_timeout(ni, TO_RELEASE, NSReleaseTimeout);
-#ifdef IRC_DAL4_4_15
+#if defined(IRC_DAL4_4_15) || defined(IRC_BAHAMUT)
     }
 #endif
 }
@@ -1224,6 +1423,34 @@ static void del_ns_timeout(NickInfo *ni, int type)
 /*********************** NickServ command routines ***********************/
 /*************************************************************************/
 
+#ifdef DEBUG_COMMANDS
+
+/* List NickServ commands */
+
+static void do_commands(User *u)
+{
+    Command *c;
+    char buf[50], *end;
+
+    end = buf;
+    *end = 0;
+    for (c = cmds; c->name; c++) {
+	if ((c->has_priv == NULL) || c->has_priv(u)) {
+	    if ((end-buf)+strlen(c->name)+2 >= sizeof(buf)) {
+		notice(s_NickServ, u->nick, "%s,", buf);
+		end = buf;
+		*end = 0;
+	    }
+	    end += snprintf(end, sizeof(buf)-(end-buf), "%s%s", 
+	    (end == buf) ? "" : ", ", c->name);
+	}
+    }
+    if (buf[0])
+	notice(s_NickServ, u->nick, "%s", buf);
+}
+
+#endif /* DEBUG_COMMANDS */
+
 /* Return a help message. */
 
 static void do_help(User *u)
@@ -1263,7 +1490,7 @@ static void do_register(User *u)
 	return;
     }
 
-#ifdef IRC_DAL4_4_15
+#if defined(IRC_DAL4_4_15) || defined(IRC_BAHAMUT)
     /* Prevent "Guest" nicks from being registered. -TheShadow */
     if (NSForceNickChange) {
 	int prefixlen = strlen(NSGuestNickPrefix);
@@ -1295,6 +1522,10 @@ static void do_register(User *u)
 	    log("%s: %s@%s tried to register FORBIDden nick %s", s_NickServ,
 			u->username, u->host, u->nick);
 	    notice_lang(s_NickServ, u, NICK_CANNOT_BE_REGISTERED, u->nick);
+	} else if (u->ni && (u->ni->flags & NI_SUSPENDED)) {
+	    log("%s: %s@%s tried to register SUSPENDED nick %s", s_NickServ,
+			u->username, u->host, u->nick);
+	    notice_lang(s_NickServ, u, NICK_CANNOT_BE_REGISTERED, u->nick);
 	} else {
 	    notice_lang(s_NickServ, u, NICK_ALREADY_REGISTERED, u->nick);
 	}
@@ -1368,8 +1599,8 @@ static void do_register(User *u)
 	    notice_lang(s_NickServ, u, NICK_PASSWORD_IS, ni->pass);
 #endif
 	    u->lastnickreg = time(NULL);
-#ifdef IRC_DAL4_4_15
-	    send_cmd(ServerName, "SVSMODE %s +r", u->nick);
+#if defined(IRC_DAL4_4_15) || defined(IRC_BAHAMUT)
+	    send_cmd(s_NickServ, "SVSMODE %s :+r", u->nick);
 #endif
 	} else {
 	    log("%s: makenick(%s) failed", s_NickServ, u->nick);
@@ -1394,6 +1625,9 @@ static void do_identify(User *u)
     } else if (!(ni = u->real_ni)) {
 	notice(s_NickServ, u->nick, "Your nick isn't registered.");
 
+    } else if (ni->flags & NI_SUSPENDED) {
+	notice_lang(s_NickServ, u, NICK_X_SUSPENDED, u->nick);
+
     } else if (!(res = check_password(pass, ni->pass))) {
 	log("%s: Failed IDENTIFY for %s!%s@%s",
 		s_NickServ, u->nick, u->username, u->host);
@@ -1416,8 +1650,8 @@ static void do_identify(User *u)
 		free(ni->last_realname);
 	    ni->last_realname = sstrdup(u->realname);
 	}
-#ifdef IRC_DAL4_4_15
-	send_cmd(ServerName, "SVSMODE %s +r", u->nick);
+#if defined(IRC_DAL4_4_15) || defined(IRC_BAHAMUT)
+	send_cmd(s_NickServ, "SVSMODE %s :+r", u->nick);
 #endif
 	log("%s: %s!%s@%s identified for nick %s", s_NickServ,
 			u->nick, u->username, u->host, u->nick);
@@ -1460,8 +1694,8 @@ static void do_drop(User *u)
     } else {
 	if (readonly)
 	    notice_lang(s_NickServ, u, READ_ONLY_MODE);
-#ifdef IRC_DAL4_4_15
-	send_cmd(ServerName, "SVSMODE %s -r", ni->nick);
+#if defined(IRC_DAL4_4_15) || defined(IRC_BAHAMUT)
+	send_cmd(s_NickServ, "SVSMODE %s :-r", ni->nick);
 #endif
 	delnick(ni);
 	log("%s: %s!%s@%s dropped nickname %s", s_NickServ,
@@ -1508,6 +1742,8 @@ static void do_set(User *u)
 	notice_lang(s_NickServ, u, MORE_INFO, s_NickServ, "SET");
     } else if (!ni) {
 	notice_lang(s_NickServ, u, NICK_NOT_REGISTERED);
+    } else if (ni->status & NS_VERBOTEN) {
+	notice_lang(s_NickServ, u, NICK_X_FORBIDDEN, ni->nick);
     } else if (!is_servadmin && !nick_identified(u)) {
 	notice_lang(s_NickServ, u, NICK_IDENTIFY_REQUIRED, s_NickServ);
     } else if (stricmp(cmd, "PASSWORD") == 0) {
@@ -1949,6 +2185,8 @@ static void do_link(User *u)
 	    ni->memos.memocount = 0;
 	}
 	u->ni = target;
+	log("%s: %s!%s@%s linked nick %s to %s", s_NickServ, u->nick,
+			u->username, u->host, u->nick, nick);
 	notice_lang(s_NickServ, u, NICK_LINKED, nick);
 	/* They gave the password, so they might as well have IDENTIFY'd */
 	target->status |= NS_IDENTIFIED;
@@ -1976,7 +2214,7 @@ static void do_unlink(User *u)
 	    syntax_error(s_NickServ, u, "UNLINK", NICK_UNLINK_SYNTAX);
 	} else if (!is_servadmin &&
 				!(res = check_password(pass, ni->pass))) {
-	    log("%s: LINK: bad password for %s by %s!%s@%s",
+	    log("%s: UNLINK: bad password for %s by %s!%s@%s",
 		s_NickServ, nick, u->nick, u->username, u->host);
 	    notice_lang(s_NickServ, u, PASSWORD_INCORRECT);
 	    bad_password(u);
@@ -1986,6 +2224,8 @@ static void do_unlink(User *u)
 	    linkname = ni->link->nick;
 	    delink(ni);
 	    notice_lang(s_NickServ, u, NICK_X_UNLINKED, ni->nick, linkname);
+	    log("%s: %s!%s@%s unlinked nick %s from %s", s_NickServ, u->nick,
+			u->username, u->host, ni->nick, linkname);
 	    /* Adjust user record if user is online */
 	    /* FIXME: probably other cases we need to consider here */
 	    for (u = firstuser(); u; u = nextuser()) {
@@ -2008,6 +2248,8 @@ static void do_unlink(User *u)
 	    u->ni = ni;  /* Effective nick now the same as real nick */
 	    delink(ni);
 	    notice_lang(s_NickServ, u, NICK_UNLINKED, linkname);
+	    log("%s: %s!%s@%s unlinked nick %s from %s", s_NickServ, u->nick,
+			u->username, u->host, u->nick, linkname);
 	}
     }
 }
@@ -2031,15 +2273,21 @@ static void do_listlinks(User *u)
 	notice_lang(s_NickServ, u, NICK_X_FORBIDDEN, ni->nick);
 
     } else {
-	notice_lang(s_NickServ, u, NICK_LISTLINKS_HEADER, ni->nick);
 	if (param)
 	    ni = getlink(ni);
+
+	notice_lang(s_NickServ, u, NICK_LISTLINKS_HEADER, ni->nick);
+
 	for (i = 0; i < 256; i++) {
 	    for (ni2 = nicklists[i]; ni2; ni2 = ni2->next) {
 		if (ni2 == ni)
 		    continue;
 		if (param ? getlink(ni2) == ni : ni2->link == ni) {
-		    notice(s_NickServ, u->nick, "    %s", ni2->nick);
+		    if (ni2->link == ni)
+			notice_lang(s_NickServ, u, NICK_X_IS_LINKED, ni2->nick);
+		    else
+			notice_lang(s_NickServ, u, NICK_X_IS_LINKED_VIA_X, 
+					ni2->nick, ni2->link->nick);
 		    count++;
 		}
 	    }
@@ -2062,7 +2310,6 @@ static void do_info(User *u)
     char *nick = strtok(NULL, " ");
     char *param = strtok(NULL, " ");
     NickInfo *ni, *real;
-    int is_servadmin = is_services_admin(u);
 
     if (!nick) {
     	syntax_error(s_NickServ, u, "INFO", NICK_INFO_SYNTAX);
@@ -2079,7 +2326,7 @@ static void do_info(User *u)
 	const char *commastr = getstring(u->ni, COMMA_SPACE);
 	int need_comma = 0;
 	int nick_online = 0;
-	int show_hidden = 0;
+	int show_all = 0;
 
 	/* Is the real owner of the nick we're looking up online? -TheShadow */
 	if (ni->status & NS_IDENTIFIED)
@@ -2090,7 +2337,7 @@ static void do_info(User *u)
         if (param && stricmp(param, "ALL") == 0 && 
 			((nick_online && (stricmp(u->nick, nick) == 0)) ||
                         	is_services_admin(u)))
-            show_hidden = 1;
+            show_all = 1;
 
 	real = getlink(ni);
 
@@ -2098,7 +2345,7 @@ static void do_info(User *u)
 		nick, ni->last_realname);
 
 	if (nick_online) {
-	    if (show_hidden || !(real->flags & NI_HIDE_MASK))
+	    if (show_all || !(real->flags & NI_HIDE_MASK))
 		notice_lang(s_NickServ, u, NICK_INFO_ADDRESS_ONLINE,
 			ni->last_usermask);
 	    else
@@ -2106,7 +2353,7 @@ static void do_info(User *u)
 			ni->nick);
 
 	} else {
-	    if (show_hidden || !(real->flags & NI_HIDE_MASK))
+	    if (show_all || !(real->flags & NI_HIDE_MASK))
 		notice_lang(s_NickServ, u, NICK_INFO_ADDRESS,
 			ni->last_usermask);
 
@@ -2118,11 +2365,11 @@ static void do_info(User *u)
 	tm = localtime(&ni->time_registered);
 	strftime_lang(buf, sizeof(buf), u, STRFTIME_DATE_TIME_FORMAT, tm);
 	notice_lang(s_NickServ, u, NICK_INFO_TIME_REGGED, buf);
-	if (ni->last_quit && (show_hidden || !(real->flags & NI_HIDE_QUIT)))
+	if (ni->last_quit && (show_all || !(real->flags & NI_HIDE_QUIT)))
 	    notice_lang(s_NickServ, u, NICK_INFO_LAST_QUIT, ni->last_quit);
 	if (ni->url)
 	    notice_lang(s_NickServ, u, NICK_INFO_URL, ni->url);
-	if (ni->email && (show_hidden || !(real->flags & NI_HIDE_EMAIL)))
+	if (ni->email && (show_all || !(real->flags & NI_HIDE_EMAIL)))
 	    notice_lang(s_NickServ, u, NICK_INFO_EMAIL, ni->email);
 	*buf = 0;
 	end = buf;
@@ -2145,35 +2392,63 @@ static void do_info(User *u)
 	}
 	notice_lang(s_NickServ, u, NICK_INFO_OPTIONS,
 		*buf ? buf : getstring(u->ni, NICK_INFO_OPT_NONE));
-	if ((ni->status & NS_NO_EXPIRE) && (real == u->ni || is_servadmin))
+
+	if (ni->link && show_all)
+	    notice_lang(s_NickServ, u, NICK_INFO_LINKED_TO, ni->link->nick);
+
+	if ((ni->status & NS_NO_EXPIRE) && show_all)
 	    notice_lang(s_NickServ, u, NICK_INFO_NO_EXPIRE);
 
+	if (real->flags & NI_SUSPENDED) {
+	    NickSuspend *snick;
+
+	    notice_lang(s_NickServ, u, NICK_X_SUSPENDED, nick);
+
+	    snick = find_nicksuspend(nick);
+	    if (show_all && !(snick = find_nicksuspend(nick))) {
+		/* FIXME: remove me someday */
+		log("ERROR: Suspend info for %s could not be found.", nick);
+		notice(s_NickServ, u->nick, "Could not find suspend info.");
+
+	    } else if (show_all) {
+		char timebuf[32], expirebuf[256];
+		time_t now = time(NULL);
+
+		tm = localtime(&snick->suspendinfo.suspended);
+		strftime_lang(timebuf, sizeof(timebuf), u, 
+				STRFTIME_DATE_TIME_FORMAT, tm);
+                if (snick->suspendinfo.expires == 0) {
+                    snprintf(expirebuf, sizeof(expirebuf),
+                                getstring(u->ni, OPER_AKILL_NO_EXPIRE));
+                } else if (snick->suspendinfo.expires <= now) {
+                    snprintf(expirebuf, sizeof(expirebuf),
+                                getstring(u->ni, OPER_AKILL_EXPIRES_SOON));
+                } else {
+		    expires_in_lang(expirebuf, sizeof(expirebuf), u,
+                                snick->suspendinfo.expires - now + 59);
+		}
+		notice_lang(s_NickServ, u, NICK_INFO_SUSPEND_DETAILS,
+				snick->suspendinfo.who, timebuf, expirebuf);
+		notice_lang(s_NickServ, u, NICK_INFO_SUSPEND_REASON, 
+				snick->suspendinfo.reason);
+	    }
+	}
+
     }
 }
 
 /*************************************************************************/
 
-/* SADMINS can search for nicks based on their NS_VERBOTEN and NS_NO_EXPIRE
- * status. The keywords FORBIDDEN and NOEXPIRE represent these two states
- * respectively. These keywords should be included after the search pattern.
- * Multiple keywords are accepted and should be separated by spaces. Only one
- * of the keywords needs to match a nick's state for the nick to be displayed.
- * Forbidden nicks can be identified by "[Forbidden]" appearing in the last
- * seen address field. Nicks with NOEXPIRE set are preceeded by a "!". Only
- * SADMINS will be shown forbidden nicks and the "!" indicator.
- * Syntax for sadmins: LIST pattern [FORBIDDEN] [NOEXPIRE]
- * -TheShadow
- */
-
 static void do_list(User *u)
 {
     char *pattern = strtok(NULL, " ");
     char *keyword;
-    NickInfo *ni;
+    NickInfo *ni, *real;
     int nnicks, i;
     char buf[BUFSIZE];
     int is_servadmin = is_services_admin(u);
-    int16 matchflags = 0; /* NS_ flags a nick must match one of to qualify */
+    int16 match_NS = 0; /* NS_ flags a nick must match one of to qualify */
+    int32 match_NI = 0; /* NI_ flags a nick must match one of to qualify */
 
     if (NSListOpersOnly && !(u->mode & UMODE_O)) {
 	notice_lang(s_NickServ, u, PERMISSION_DENIED);
@@ -2188,19 +2463,24 @@ static void do_list(User *u)
 
 	while (is_servadmin && (keyword = strtok(NULL, " "))) {
 	    if (stricmp(keyword, "FORBIDDEN") == 0)
-		matchflags |= NS_VERBOTEN;
+		match_NS |= NS_VERBOTEN;
 	    if (stricmp(keyword, "NOEXPIRE") == 0)
-		matchflags |= NS_NO_EXPIRE;
+		match_NS |= NS_NO_EXPIRE;
+	    if (stricmp(keyword, "SUSPENDED") == 0)
+		match_NI |= NI_SUSPENDED;
 	}
 
 	notice_lang(s_NickServ, u, NICK_LIST_HEADER, pattern);
 	for (i = 0; i < 256; i++) {
 	    for (ni = nicklists[i]; ni; ni = ni->next) {
+		real = getlink(ni);
 		if (!is_servadmin && ((ni->flags & NI_PRIVATE)
 						|| (ni->status & NS_VERBOTEN)))
 		    continue;
-		if ((matchflags != 0) && !(ni->status & matchflags))
-		    continue;
+		if ((match_NI | match_NS) != 0)
+		    /* ok, we have flags, now see if they match */
+		    if (!((ni->status & match_NS) || (real->flags & match_NI)))
+			continue;
 
 		/* We no longer compare the pattern against the output buffer.
 		 * Instead we build a nice nick!user@host buffer to compare.
@@ -2210,9 +2490,16 @@ static void do_list(User *u)
 		if (stricmp(pattern, ni->nick) == 0 ||
 					match_wild_nocase(pattern, buf)) {
 		    if (++nnicks <= NSListMax) {
+			/* This is becoming unpretty. Ideas? -TheShadow */
 			char noexpire_char = ' ';
-			if (is_servadmin && (ni->status & NS_NO_EXPIRE))
-			    noexpire_char = '!';
+			char suspended_char = ' ';
+
+			if (is_servadmin) {
+			    if (ni->status & NS_NO_EXPIRE)
+				noexpire_char = '!';
+			    if (real->flags & NI_SUSPENDED)
+				suspended_char = '*';
+			}
 			if (!is_servadmin && (ni->flags & NI_HIDE_MASK)) {
 			    snprintf(buf, sizeof(buf), "%-20s  [Hidden]",
 						ni->nick);
@@ -2223,8 +2510,8 @@ static void do_list(User *u)
 			    snprintf(buf, sizeof(buf), "%-20s  %s",
 						ni->nick, ni->last_usermask);
 			}
-			notice(s_NickServ, u->nick, "   %c%s",
-						noexpire_char, buf);
+			notice(s_NickServ, u->nick, "   %c%c %s",
+					suspended_char, noexpire_char, buf);
 		    }
 		}
 	    }
@@ -2417,10 +2704,17 @@ static void do_forbid(User *u)
 	syntax_error(s_NickServ, u, "FORBID", NICK_FORBID_SYNTAX);
 	return;
     }
+    if ((ni = findnick(nick)) != NULL) {
+	if (NSSecureAdmins && nick_is_services_admin(ni) &&
+                                                        !is_services_root(u)) {
+	    notice_lang(s_NickServ, u, PERMISSION_DENIED);
+	    return;
+	}
+	delnick(ni);
+    }
+
     if (readonly)
 	notice_lang(s_NickServ, u, READ_ONLY_MODE);
-    if ((ni = findnick(nick)) != NULL)
-	delnick(ni);
     ni = makenick(nick);
     if (ni) {
 	ni->status |= NS_VERBOTEN;
@@ -2434,3 +2728,96 @@ static void do_forbid(User *u)
 }
 
 /*************************************************************************/
+
+static void do_suspend(User *u)
+{
+    NickInfo *ni, *real;
+    NickSuspend *snick;
+    char *expiry, *nick, *reason;
+    time_t expires;
+
+    /* Assumes that permission checking has already been done. */
+
+    nick = strtok(NULL, " ");
+    if (nick && *nick == '+') {
+	expiry = nick;
+	nick = strtok(NULL, " ");
+    } else {
+	expiry = NULL;
+    }
+
+    reason = strtok(NULL, "");
+
+    if (!reason) {
+	syntax_error(s_NickServ, u, "SUSPEND", NICK_SUSPEND_SYNTAX);
+
+    } else if (!(ni = findnick(nick))) {
+	notice_lang(s_NickServ, u, NICK_X_NOT_REGISTERED, nick);
+	
+    } else if (ni->status & NS_VERBOTEN) {
+	notice_lang(s_NickServ, u, NICK_X_FORBIDDEN, nick);
+	
+    } else if ((real = getlink(ni))->flags & NI_SUSPENDED) {
+	notice_lang(s_NickServ, u, NICK_SUSPEND_ALREADY_SUSPENDED, nick);
+
+    } else {
+	if (readonly)
+	    notice_lang(s_NickServ, u, READ_ONLY_MODE);
+
+	/* FIXME: add config option for default expiry */
+	if (expiry) {
+	    expires = dotime(expiry);
+	    if (expires < 0) {
+		notice_lang(s_NickServ, u, BAD_EXPIRY_TIME);
+		return;
+	    } else if (expires > 0) {
+		expires += time(NULL);	/* Set an absolute time */
+	    }
+	} else {
+	    expires = time(NULL) + NSSuspendExpire;
+	}
+
+	real->flags |= NI_SUSPENDED;
+
+	snick = add_nicksuspend(ni, reason, u->nick, expires);
+	snick->ni = real;
+
+	notice_lang(s_NickServ, u, NICK_SUSPEND_SUCCEEDED, nick);
+    }
+}
+
+static void do_unsuspend(User *u)
+{
+    NickInfo *ni, *real;
+    NickSuspend *ns;
+    char *nick = strtok(NULL, " ");
+
+    /* Assumes that permission checking has already been done. */
+    if (!nick) {
+	syntax_error(s_NickServ, u, "UNSUSPEND", NICK_UNSUSPEND_SYNTAX);
+
+    } else if (!(ni = findnick(nick))) {
+	notice_lang(s_NickServ, u, NICK_X_NOT_REGISTERED, nick);
+	
+    } else if (ni->status & NS_VERBOTEN) {
+	notice_lang(s_NickServ, u, NICK_X_FORBIDDEN, nick);
+
+    } else if (!((real = getlink(ni))->flags & NI_SUSPENDED)) {
+	notice_lang(s_NickServ, u, NICK_SUSPEND_NOT_SUSPENDED, nick);
+	
+    } else {
+	if (readonly)
+	    notice_lang(s_NickServ, u, READ_ONLY_MODE);
+
+	if (!(ns = find_nicksuspend(nick))) {
+	    log("%s: couldn't find suspend info for %s", s_NickServ,real->nick);
+	} else {
+	    del_nicksuspend(ns);
+	    log("%s: %s UNSUSPENDED %s", s_NickServ, u->nick, real->nick);
+	}
+	unsuspend(real);
+	notice_lang(s_NickServ, u, NICK_UNSUSPEND_SUCCEEDED, nick);
+    }
+}
+
+/*************************************************************************/
