@@ -1,7 +1,9 @@
 /* Routines to maintain a list of online users.
  *
- * Services is copyright (c) 1996-1999 Andy Church.
+ * Services is copyright (c) 1996-1999 Andrew Church.
  *     E-mail: <achurch@dragonfire.net>
+ * Services is copyright (c) 1999-2000 Andrew Kempe.
+ *     E-mail: <theshadow@shadowfire.org>
  * This program is free but copyrighted software; see the file COPYING for
  * details.
  */
@@ -89,16 +91,20 @@ static void delete_user(User *user)
 
     if (debug >= 2)
 	log("debug: delete_user() called");
+
     usercnt--;
     if (user->mode & UMODE_O)
 	opcnt--;
+#ifdef STATISTICS
+    stats_do_quit(user);
+#endif
+
     cancel_user(user);
     if (debug >= 2)
 	log("debug: delete_user(): free user data");
     free(user->username);
     free(user->host);
     free(user->realname);
-    free(user->server);
     if (debug >= 2)
 	log("debug: delete_user(): remove from channels");
     c = user->chans;
@@ -154,8 +160,6 @@ void get_user_stats(long *nusers, long *memuse)
 		mem += strlen(user->host)+1;
 	    if (user->realname)
 		mem += strlen(user->realname)+1;
-	    if (user->server)
-		mem += strlen(user->server)+1;
 	    for (uc = user->chans; uc; uc = uc->next)
 		mem += sizeof(*uc);
 	    for (uci = user->founder_chans; uci; uci = uci->next)
@@ -186,17 +190,18 @@ void send_user_list(User *user)
 		u->nick, u->username, u->host,
 		(u->mode&UMODE_G)?"g":"", (u->mode&UMODE_I)?"i":"",
 		(u->mode&UMODE_O)?"o":"", (u->mode&UMODE_S)?"s":"",
-		(u->mode&UMODE_W)?"w":"", u->signon, u->server, u->realname);
+		(u->mode&UMODE_W)?"w":"", u->signon, u->server->name, 
+		u->realname);
 	buf[0] = 0;
 	s = buf;
 	for (c = u->chans; c; c = c->next)
 	    s += snprintf(s, sizeof(buf)-(s-buf), " %s", c->chan->name);
-	notice(s_OperServ, source, "%s%s", u->nick, buf);
+	notice(s_OperServ, source, "CHANNELS: %s", buf);
 	buf[0] = 0;
 	s = buf;
 	for (ci = u->founder_chans; ci; ci = ci->next)
 	    s += snprintf(s, sizeof(buf)-(s-buf), " %s", ci->chan->name);
-	notice(s_OperServ, source, "%s%s", u->nick, buf);
+	notice(s_OperServ, source, "FOUNDER: %s", buf);
     }
 }
 
@@ -222,7 +227,8 @@ void send_user_info(User *user)
 		u->nick, u->username, u->host,
 		(u->mode&UMODE_G)?"g":"", (u->mode&UMODE_I)?"i":"",
 		(u->mode&UMODE_O)?"o":"", (u->mode&UMODE_S)?"s":"",
-		(u->mode&UMODE_W)?"w":"", u->signon, u->server, u->realname);
+		(u->mode&UMODE_W)?"w":"", u->signon, u->server->name, 
+		u->realname);
     buf[0] = 0;
     s = buf;
     for (c = u->chans; c; c = c->next)
@@ -294,6 +300,15 @@ User *nextuser(void)
  *	av[0] = nick
  *	If a new user:
  *		av[1] = hop count
+ *	    For Bahamut:
+ *		av[2] = TS (time of signon or last nick change)
+ *		av[3] = user's modes (or just '+' for none)
+ *		av[4] = username
+ *		av[5] = hostname
+ *		av[6] = user's server
+ *		av[7] = services stamp (set with mode +d. 0 == no stamp)
+ *		av[8] = user's real name
+ *	    Else:
  *		av[2] = signon time
  *		av[3] = username
  *		av[4] = hostname
@@ -313,6 +328,10 @@ void do_nick(const char *source, int ac, char **av)
     if (!*source) {
 	/* This is a new user; create a User structure for it. */
 
+#ifdef IRC_BAHAMUT
+	char **av_umode;
+#endif
+
 	if (debug)
 	    log("debug: new user: %s", av[0]);
 
@@ -324,22 +343,45 @@ void do_nick(const char *source, int ac, char **av)
 	 */
 
 	/* First check for AKILLs. */
+#ifdef IRC_BAHAMUT
+	if (check_akill(av[0], av[4], av[5]))
+#else
 	if (check_akill(av[0], av[3], av[4]))
+#endif
 	    return;
 
 #ifndef STREAMLINED
         /* Now check for session limits */
+# ifdef IRC_BAHAMUT
+        if (LimitSessions && !add_session(av[0], av[5]))
+# else
         if (LimitSessions && !add_session(av[0], av[4]))
+# endif /* IRC_BAHAMUT */
             return;
-#endif
+#endif /* !STREAMLINED */
 
 	/* Allocate User structure and fill it in. */
 	user = new_user(av[0]);
 	user->signon = atol(av[2]);
+#ifdef IRC_BAHAMUT
+	user->username = sstrdup(av[4]);
+	user->host = sstrdup(av[5]);
+	user->server = findserver(av[6]);
+	user->realname = sstrdup(av[8]);
+
+	/* Set usermodes for the new user */
+	av_umode = smalloc(sizeof(char *) * 2);
+	av_umode[0] = av[0];
+	av_umode[1] = av[3];
+	do_umode(av[0], 2, av_umode);
+	free(av_umode);
+#else
 	user->username = sstrdup(av[3]);
 	user->host = sstrdup(av[4]);
-	user->server = sstrdup(av[5]);
+	user->server = findserver(av[5]);
 	user->realname = sstrdup(av[6]);
+#endif /* IRC_BAHAMUT */
+	user->timestamp = user->signon;
 	user->my_signon = time(NULL);
 
 	if (CheckClones) {
@@ -347,6 +389,10 @@ void do_nick(const char *source, int ac, char **av)
 	    check_clones(user);
 	}
 
+#ifdef STATISTICS
+	user->server->stats->usercnt++;
+#endif
+
 	display_news(user, NEWS_LOGON);
 
     } else {
@@ -366,6 +412,8 @@ void do_nick(const char *source, int ac, char **av)
 	if (stricmp(av[0], user->nick) != 0)
 	    user->my_signon = time(NULL);
 
+	user->timestamp = atol(av[1]);
+
 	new_ni = findnick(av[0]);
 	if (new_ni)
 	    new_ni = getlink(new_ni);
@@ -376,14 +424,18 @@ void do_nick(const char *source, int ac, char **av)
 	change_user_nick(user, av[0]);
     }
 
-    if (ni_changed) {
-	if (validate_user(user))
-	    check_memos(user);
-#ifdef IRC_DAL4_4_15
+    if (ni_changed && validate_user(user)) {
+#if defined(IRC_DAL4_4_15) || defined(IRC_BAHAMUT)
 	if (nick_identified(user)) {
-	    send_cmd(ServerName, "SVSMODE %s +r", av[0]);
-	}
-#endif
+	    send_cmd(s_NickServ, "SVSMODE %s :+r", av[0]);
+	} 
+#endif /* IRC_DAL4_4_15 || IRC_BAHAMUT */
+	check_memos(user);
+#ifdef IRC_DAL4_4_15
+    /* Bahamut removes the "+r" usermode automatically. */
+    } else {
+	send_cmd(s_NickServ, "SVSMODE %s :-r", av[0]);
+#endif /* IRC_DAL4_4_15 */
     }
 }
 
@@ -397,7 +449,7 @@ void do_join(const char *source, int ac, char **av)
 {
     User *user;
     char *s, *t;
-    struct u_chanlist *c;
+    struct u_chanlist *c, *nextc;
     ChannelInfo *ci;
 
     user = finduser(source);
@@ -413,6 +465,19 @@ void do_join(const char *source, int ac, char **av)
 	    *t++ = 0;
 	if (debug)
 	    log("debug: %s joins %s", source, s);
+
+        if (*s == '0') {
+            c = user->chans;
+            while (c) {
+                nextc = c->next;
+                chan_deluser(user, c->chan);
+                free(c);
+                c = nextc;
+            }
+            user->chans = NULL;
+            continue;
+        }
+
 	/* Make sure check_kick comes before chan_adduser, so banned users
 	 * don't get to see things like channel keys. */
 	if (check_kick(user, s))
@@ -432,6 +497,124 @@ void do_join(const char *source, int ac, char **av)
 
 /*************************************************************************/
 
+#ifdef IRC_BAHAMUT
+
+/* Handle an SJOIN command.
+ *	av[0] = TS3 timestamp
+ *	av[1] = TS3 timestamp - channel creation time
+ *	av[2] = channel
+ *	av[3] = channel modes
+ *	av[4] = limit / key (depends on modes in av[3])
+ *	av[5] = limit / key (depends on modes in av[3])
+ *	av[4|5|6] = nickname(s), with modes, joining channel
+ *
+ * I'm almost 100% sure this code is not fully optimised. However, it
+ * functions and that is all I'm looking to achieve right now. It should be
+ * noted that there will be problems if unknown channel modes for users
+ * are encountered. -TheShadow
+ */
+
+void do_sjoin(const char *source, int ac, char **av)
+{
+    User *user;
+    char *t, *nick;
+    char *channel = av[2];
+    char **av_cmode;
+    struct u_chanlist *c;
+    ChannelInfo *ci = NULL;
+    int modecnt;	/* Number of modes for a user (+o and/or +v) */
+
+    t = av[ac-1];	/* The nicknames are always the last param */
+    while (*(nick=t)) {
+	char modebuf[4] = { 0 };	/* Used to hold a nice mode string */
+
+        t = nick + strcspn(nick, " ");
+	if (*t)
+	    *t++ = 0;
+
+	modecnt = 0;
+	modebuf[0] = '+';
+	while (*nick == '@' || *nick == '+') {
+	    switch (*nick) {
+	      case '@':	
+		modecnt++;
+		modebuf[modecnt] = 'o';
+		break;
+	      case '+':
+		modecnt++;
+		modebuf[modecnt] = 'v';
+		break;
+	    }
+	    nick++;
+	}
+
+	user = finduser(nick);
+	if (!user) {
+	    log("SJOIN to channel %s for non-existant nick %s (%s)", 
+	    		channel, nick, merge_args(ac, av));
+	    continue;
+	}
+
+	if (debug)
+	    log("debug: %s SJOINs %s", nick, channel);
+	
+	/* Make sure check_kick comes before chan_adduser, so banned users
+	 * don't get to see things like channel keys. */
+	if (check_kick(user, channel))
+	    continue;
+	chan_adduser(user, channel);
+	if ((ci || (ci = cs_findchan(channel))) && ci->entry_message)
+	    notice(s_ChanServ, user->nick, "%s", ci->entry_message);
+	c = smalloc(sizeof(*channel));
+	c->next = user->chans;
+	c->prev = NULL;
+	if (user->chans)
+	    user->chans->prev = c;
+	user->chans = c;
+	c->chan = findchan(channel);
+
+	if (modecnt > 0) {
+	    /* Set channel modes for user.
+	     * We need to watch out for +ov and send two nicks when it 
+	     * happens. */
+
+	    if (debug)
+		log("debug: channel modes for %s are %s", nick, modebuf);
+		
+	    av_cmode = smalloc(sizeof(char *) * (modecnt + 2));
+	    av_cmode[0] = channel;
+	    av_cmode[1] = modebuf;
+	    av_cmode[2] = nick;
+	    if (modecnt == 2)
+		av_cmode[3] = nick;
+	    do_cmode(source, modecnt+2, av_cmode);
+	    free(av_cmode);
+	}
+    }
+
+    /* Did anyone actually join the channel and are there really any modes? */
+    if (ci && av[3]+1 != '\0') {
+	/* Set channel modes.
+	 * We need to watch out for the additional params for +k and +l. */
+	av_cmode = smalloc(sizeof(char *) * (ac - 3));
+	av_cmode[0] = channel;
+	av_cmode[1] = av[3];	/* The actual channel modes */
+	/* Now see if we have additional params for +k and/or +l */
+	switch (ac) {
+	  case 7:
+	    av_cmode[3] = av[5];
+	  case 6:
+	    av_cmode[2] = av[4];
+	} 
+	do_cmode(av[0], ac-3, av_cmode);
+	free(av_cmode);
+    }
+}
+
+#endif /* IRC_BAHAMUT */
+
+/*************************************************************************/
+
 /* Handle a PART command.
  *	av[0] = channels to leave
  *	av[1] = reason (optional)
@@ -551,12 +734,19 @@ void do_umode(const char *source, int ac, char **av)
 	switch (*s++) {
 	    case '+': add = 1; break;
 	    case '-': add = 0; break;
-#ifdef IRC_DAL4_4_15
+#if defined(IRC_DAL4_4_15) || defined(IRC_BAHAMUT)
 	    case 'r': 
 	    	if (add && !nick_identified(user)) {
-		    send_cmd(ServerName, "SVSMODE %s -r", av[0]);
+		    send_cmd(s_NickServ, "SVSMODE %s :-r", av[0]);
 		}
 		break;
+#endif
+#ifdef IRC_BAHAMUT
+	    case 'd': 
+	    	log("user: MODE %s from %s for %s", av[1], source, av[0]);
+		wallops(NULL, "%s tried to set mode %s for %s", source, av[1],
+				av[0]);
+		break;
 #endif
 	    case 'i': add ? (user->mode |= UMODE_I) : (user->mode &= ~UMODE_I);
 	              break;
@@ -579,6 +769,10 @@ void do_umode(const char *source, int ac, char **av)
 		    user->mode &= ~UMODE_O;
 		    opcnt--;
 		}
+#ifdef STATISTICS
+		add ? user->server->stats->opercnt++ 
+				: user->server->stats->opercnt--;
+#endif
 		break;
 	}
     }
